use log::{debug, info};

use mio::net::TcpStream;
use mio::event::Event;
use mio::{Interest, Registry};

use std::io::{self, Read, Write};

use crate::state::State;
use crate::system::System;
use crate::user::User;
use crate::utils::{would_block, interrupted};

pub struct Connection<'a> {
    pub system: &'a System,
    pub ts: TcpStream,
    pub user: Option<User>,
    pub state: State,
}

impl Connection<'_> {
    pub fn handle_event(&mut self, reg: &Registry, event: &Event) -> io::Result<bool> {
        if event.is_writable() {
            self.handle_write_event(reg, event)?;
        }

        if event.is_readable() {
            self.handle_read_event(reg, event)?;
        }

        Ok(false)
    }

    fn set_state(&mut self, s: State) {
        debug!("state change {:?} -> {:?}", self.state, s);
        self.state = s;
    }

    fn write(&mut self, buf: &[u8]) -> io::Result<bool> {
        loop {
            match self.ts.write(&buf) {
                // This shouldn't happen since the TCP send buffer is big enough
                Ok(n) if n < buf.len() => return Err(io::ErrorKind::WriteZero.into()),
                Ok(_) => return Ok(true),
                Err(ref err) if would_block(err) => return Ok(false),
                Err(ref err) if interrupted(err) => continue,
                Err(err) => return Err(err),
            }
        }
    }

    fn handle_write_event(&mut self, reg: &Registry, event: &Event) -> io::Result<()> {
        // TODO just write everything that changed in the framebuffer
        // we update the fb after user input, there is no change until the user
        // did something (for now), so we don't need to match on the current state here
        // a special case might be after the user logged out, but this could also be
        // represented as a new state
        // ... actually... this works only after the user has seen the banner since
        // it scrolls the screen, so we still need to match the state but only check
        // for the Connected and maybe the Logout state, everything else just prints
        // updates of the fb
        match self.state {
            // query term info (size etc.)
            State::Connected => {
                if self.write(&self.system.theme.banner)? {
                    self.set_state(State::Banner);
                    reg.reregister(&mut self.ts, event.token(), Interest::READABLE)?;
                }
            },
            // send banner and wait for keypress
            _ => todo!("handle write for all the other states")
        };

        Ok(())
    }

    fn read(&self, buf: &mut [u8]) -> io::Result<()> {

        Ok(())
    }

    fn handle_read_event(&mut self, reg: &Registry, event: &Event) -> io::Result<bool> {
        let mut received_data = vec![0; 4096];
        let mut connection_closed = false;
        let mut bytes_read = 0;


        loop {
            match self.ts.read(&mut received_data[bytes_read..]) {
                Ok(0) => {
                    connection_closed = true;
                    break;
                }
                Ok(n) => {
                    bytes_read += n;
                    // TODO: we should probably fail when our recv buffer is too
                    // small and not just increase it indefinitely
                    if bytes_read == received_data.len() {
                        received_data.resize(received_data.len() + 1024, 0);
                    }
                }
                Err(ref err) if would_block(err) => break,
                Err(ref err) if interrupted(err) => continue,
                Err(err) => return Err(err),
            }
        }


        if bytes_read != 0 {
            let received_data = &received_data[..bytes_read];
            debug!("data: {:?}", received_data);

            match self.state {
                // parse term info
                State::Connected => {

                },
                // one keypress to show the login screen, ignore the rest
                State::Banner => {},
                _ => {},
            };

            // if let Ok(str_buf) = std::str::from_utf8(received_data) {
            //     debug!("Received data: {}", str_buf.trim_end());
            // } else {
            //     debug!("Received (none UTF-8) data: {:?}", received_data);
            // }
        }

        if connection_closed {
            info!("Connection closed");
            return Ok(true);
        }

        Ok(false)
    }
}
