use log::info;
use mio::net::TcpListener;
use mio::{Events, Interest, Poll, Token};
use std::collections::HashMap;
use std::io;

mod connection;
mod system;
mod state;
mod theme;
mod user;
mod utils;

use self::connection::Connection;
use self::system::System;
use self::state::State;
use self::utils::interrupted;

const SERVER: Token = Token(0);
const LISTEN: &str = "127.0.0.1:9000";

fn main() -> io::Result<()> {
    env_logger::init();

    let system = System::new("dachgeschoss")?;
    let mut poll = Poll::new()?;
    let mut events = Events::with_capacity(128);

    let addr = LISTEN.parse().unwrap();
    let mut server = TcpListener::bind(addr)?;

    poll.registry()
        .register(&mut server, SERVER, Interest::READABLE)?;

    let mut connections = HashMap::new();
    let mut unique_token = Token(SERVER.0 + 1);

    loop {
        if let Err(err) = poll.poll(&mut events, None) {
            if interrupted(&err) {
                continue;
            }
            return Err(err);
        }

        for event in events.iter() {
            match event.token() {
                SERVER => loop {
                    let (mut connection, address) = match server.accept() {
                        Ok((connection, address)) => (connection, address),
                        Err(e) if e.kind() == io::ErrorKind::WouldBlock => {
                            break;
                        }
                        Err(e) => {
                            return Err(e);
                        }
                    };

                    info!("Connection from: {}", address);

                    let token = next(&mut unique_token);
                    poll.registry().register(
                        &mut connection,
                        token,
                        Interest::READABLE.add(Interest::WRITABLE),
                    )?;

                    connections.insert(
                        token,
                        Connection {
                            system: &system,
                            ts: connection,
                            state: State::Connected,
                            user: None,
                        },
                    );
                },
                token => {
                    let done = if let Some(connection) = connections.get_mut(&token) {
                        connection.handle_event(poll.registry(), event)?
                    } else {
                        false
                    };

                    if done {
                        if let Some(mut connection) = connections.remove(&token) {
                            poll.registry().deregister(&mut connection.ts)?;
                        }
                    }
                }
            }
        }
    }
}

fn next(current: &mut Token) -> Token {
    let next = current.0;
    current.0 += 1;
    Token(next)
}

